proc todo_init_var {org_id proj_id todo_id}\
{
	global TODO_REC_${org_id}_${proj_id}_${todo_id}
	set TODO_REC_${org_id}_${proj_id}_${todo_id} ""
}

proc todo_new_id {org_id proj_id}\
{
	global LIST_TODO
	set todo_id 0
	foreach b $LIST_TODO \
	{
		if {[regexp "^${org_id}_${proj_id}_(\[0-9\]+)" $b all id]}\
		{
			if {$id > $todo_id}\
			{
				set todo_id $id
			}
		}
	}
	return [expr $todo_id + 1]
}

proc todo_d_new {org_id}\
{
	global proj
	global CUR_ORG LIST_PROJ LIST_TODO
	catch {destroy .newtodo}
	toplevel .newtodo
	wm title .newtodo "New todo list"
	wm geometry .newtodo +50+50
	frame .newtodo.b
	pack .newtodo.b -side bottom -pady 5 -padx 5
		button .newtodo.b.ok -text Ok -width 10 \
			-command {
				if {[.newtodo.l.f3.lb size] == 0}\
				{
					return
				}
				set proj_id [join [lindex [split $proj ---] 0]]
				set todo_id [todo_new_id $CUR_ORG $proj_id]
				todo_init_var $CUR_ORG $proj_id $todo_id
				upvar #0 TODO_REC_${CUR_ORG}_${proj_id}_${todo_id} TODO_REC
				set TODO_REC ""
				upvar #0 TODO_${CUR_ORG}_${proj_id}_${todo_id} TODO
				set TODO ""
				if {[join [.newtodo.name.e get]] == ""}\
				{
					keylset TODO name "Untitled"
				}\
				else\
				{
					keylset TODO name [.newtodo.name.e get]
				}
				lappend LIST_TODO ${CUR_ORG}_${proj_id}_${todo_id}
				todo_save $CUR_ORG $proj_id $todo_id
				o_menu_compose $CUR_ORG
				destroy .newtodo
			}
		pack .newtodo.b.ok -side left
		button .newtodo.b.cancel -text Cancel -width 10 \
			-command "
				destroy .newtodo
			"
		pack .newtodo.b.cancel -side left
	frame .newtodo.name
	pack .newtodo.name -fill x
		label .newtodo.name.l -text "Todo list's name: "
		pack .newtodo.name.l -side left
		entry .newtodo.name.e
		pack .newtodo.name.e -side left
	frame .newtodo.l
	pack .newtodo.l -side left -fill y -pady 5 -padx 5
		frame .newtodo.l.f3
		pack .newtodo.l.f3
			scrollbar .newtodo.l.f3.vs -command {.newtodo.l.f3.lb yview}
			pack .newtodo.l.f3.vs -side right -fill y
			listbox .newtodo.l.f3.lb -width 30 -yscrollcommand {.newtodo.l.f3.vs set}
			pack .newtodo.l.f3.lb -side right -fill y
			foreach p $LIST_PROJ \
			{
				global PROJ_${org_id}_$p
				.newtodo.l.f3.lb insert end "$p --- [keylget PROJ_${org_id}_$p name]"
			}
			.newtodo.l.f3.lb selection set 0
			set proj [.newtodo.l.f3.lb get 0]
			bind .newtodo.l.f3.lb <1> {
				set proj [.newtodo.l.f3.lb get @%x,%y]
			}
	focus .newtodo.name.e
}

proc todo_save {org_id proj_id todo_id}\
{
	global ORG_PATH
	upvar #0 TODO_${org_id}_${proj_id}_${todo_id} TODO
	upvar #0 TODO_REC_${org_id}_${proj_id}_${todo_id} TODO_REC
	set ct [catch {
		set F [open $ORG_PATH/org_${org_id}/proj_${org_id}_${proj_id}/todo_${org_id}_${proj_id}_${todo_id}.tmp w+]
	} errmsg]
	if {$ct > 0}\
	{
		myError $errmsg
		return
	}
	set ct [catch {
	puts $F "set TODO_${org_id}_${proj_id}_${todo_id} \"\""
	puts $F "\n# STRUCTURE"
	puts $F "keylset TODO_${org_id}_${proj_id}_${todo_id} name \"[keylget TODO name]\""
	puts $F "\n# RECORDS"
	puts $F "set TODO_REC_${org_id}_${proj_id}_${todo_id} \"\""
	foreach i [keylkeys TODO_REC] \
	{
		puts $F "keylset TODO_REC_${org_id}_${proj_id}_${todo_id} $i [list [keylget TODO_REC $i]]"
	}
	close $F
	} errmsg]
	if {$ct > 0}\
	{
		unlink $ORG_PATH/org_${org_id}/proj_${org_id}_${proj_id}/todo_${org_id}_${proj_id}_${todo_id}.tmp
		myError "Cannot save todo. Reason:\n$errmsg"
	}\
	else\
	{
		frename $ORG_PATH/org_${org_id}/proj_${org_id}_${proj_id}/todo_${org_id}_${proj_id}_${todo_id}.tmp $ORG_PATH/org_${org_id}/proj_${org_id}_${proj_id}/todo_${org_id}_${proj_id}_${todo_id}
		chmod 0600 $ORG_PATH/org_${org_id}/proj_${org_id}_${proj_id}/todo_${org_id}_${proj_id}_${todo_id}
	}
}

proc todo_export_ascii {org_id proj_id todo_id}\
{
	upvar #0 TODO_${org_id}_${proj_id}_${todo_id} TODO
	upvar #0 TODO_REC_${org_id}_${proj_id}_${todo_id} TODO_REC
	global FILENAME TASK_SORT
	o_wait
	todo_save $org_id $proj_id $todo_id
	set ct [catch {
		set F [open [pwd]/$FILENAME w+]
	} errmsg]
	if {$ct > 0}\
	{
		myError "Cannot save the todo list as text. Reason:\n$errmsg"
		return
	}
	puts $F "[keylget TODO name]\n"
	puts $F "+--------+--------------------------------------------------+-----+"
	set line [format "|%8s|%-50.50s|%5s|" "Priority" "Description" "State"]
	puts $F "$line"
	puts $F "+--------+--------------------------------------------------+-----+"
	foreach i [lsort -command "todo_sort $org_id $proj_id $todo_id $TODO_SORT" [keylkeys TODO_REC]] \
	{
		set line [format "|%8s|%-50.50s|%5s|" [replicate "*" [keylget TODO_REC $i.priority]] [keylget TODO_REC $i.description] [keylget TODO_REC $i.state]]
		puts $F "$line"
	}
	puts $F "+--------+--------------------------------------------------+-----+"
	close $F
	o_idle
}

proc todo_export_html {org_id proj_id todo_id}\
{
	upvar #0 TODO_${org_id}_${proj_id}_${todo_id} TODO
	upvar #0 TODO_REC_${org_id}_${proj_id}_${todo_id} TODO_REC
	global FILENAME TODO_COL TODO_SORT
	o_wait
	todo_save $org_id $proj_id $todo_id
	if {[file extension $FILENAME] != ".html"}\
	{
		append FILENAME ".html"
	}
	set ct [catch {
		set F [open [pwd]/$FILENAME w+]
	} errmsg]
	if {$ct > 0}\
	{
		myError "Cannot save the todo list as text. Reason:\n$errmsg"
		return
	}
	puts $F "<html><head>"
	puts $F "<title>[html_text2html [keylget TODO name]]</title></head>"
	puts $F "<body bgcolor=\"#FFFFFF\">"
	puts $F "<h2>[html_text2html [keylget TODO name]]</h2>"
	puts $F "<table>"
	puts $F "\t<tr><td><b>Description</b></td><td><b>State</b></td></tr>"
	foreach i [lsort -command "todo_sort $org_id $proj_id $todo_id $TODO_SORT" [keylkeys TODO_REC]] \
	{
		puts $F "\t<tr><td bgcolor=\"$TODO_COL([keylget TODO_REC $i.priority])\">[html_text2html [keylget TODO_REC $i.description]]</td><td bgcolor=\"$TODO_COL([keylget TODO_REC $i.priority])\">[keylget TODO_REC $i.state]</td></tr>"
	}
	puts $F "</table>"
	puts $F "</body></html>"
	close $F
	o_idle
}

proc todo_d_export {org_id proj_id todo_id mode}\
{
	global FILENAME fbe 
	catch {destroy .file}
	toplevel .file
	wm title .file "Export todo list - $mode"
	wm geometry .file 250x250+100+50
	filebrowser .file.fb 
	pack .file.fb
		if {$mode == "html"}\
		{
			.file.fb addlistbox .*.html -selectmode single
		}\
		else\
		{
			.file.fb addlistbox .* -selectmode single
		}
		.file.fb addlabel -text "Export: "
		.file.fb adddirlabel
		set fbe [.file.fb addentry -width 20]
		.file.fb addbutton -text OK -width 10 \
			-command "
				set FILENAME \[$fbe get\]
				if {\$FILENAME != \"\"}\
				{
					destroy .file
					todo_export_$mode $org_id $proj_id $todo_id
				}
			"
		.file.fb addbutton -text Cancel -width 10 \
			-command "destroy .file"
		if {$mode == "html"}\
		{
			.file.fb readdir [pwd] .*.html
		}\
		else\
		{
			.file.fb readdir [pwd] .*
		}
	focus $fbe
}

proc todo_delete {org_id proj_id todo_id}\
{
	global ORG_PATH LIST_TODO CUR_TODO CUR_TYPE
	global TODO_${org_id}_${proj_id}_${todo_id}
	global TODO_REC_${org_id}_${proj_id}_${todo_id}
	global confirm
	gen_d_yn_confirm "Do you really want to delete \"[keylget TODO_${org_id}_${proj_id}_${todo_id} name]\"?"
	tkwait variable confirm
	if {$confirm == "yes"}\
	{
		unlink $ORG_PATH/org_${org_id}/proj_${org_id}_${proj_id}/todo_${org_id}_${proj_id}_${todo_id}
		set LIST_TODO [lindex [intersect3 $LIST_TODO ${org_id}_${proj_id}_${todo_id}] 0]
		unset TODO_${org_id}_${proj_id}_${todo_id}
		unset TODO_REC_${org_id}_${proj_id}_${todo_id}
		o_menu_compose $org_id
		set CUR_TYPE proj
		o_display_proj $org_id $proj_id
	}
}

##############################

proc todo_sort {org_id proj_id todo_id field id1 id2}\
{
	upvar #0 TODO_REC_${org_id}_${proj_id}_${todo_id} TODO_REC
	if {$field == "id"}\
	{
		if {$id1 > $id2}\
		{
			return 1
		}\
		elseif {$id1 < $id2}\
		{
			return -1
		}\
		else\
		{
			return 0
		}
	}\
	else\
	{
		return [ccollate [keylget TODO_REC $id1.$field] [keylget TODO_REC $id2.$field]]
	}
}

proc todo_d_ask_description {org_id proj_id todo_id id}\
{
	upvar #0 TODO_${org_id}_${proj_id}_${todo_id} TODO
	upvar #0 TODO_REC_${org_id}_${proj_id}_${todo_id} TODO_REC
	global description state priority TODO_COL
	catch {destroy .ask}
	toplevel .ask
	wm title .ask "Description"
	wm geometry .ask +100+100
	if {[keylget TODO_REC $id.state {}]}\
	{
		set state [keylget TODO_REC $id.state]
	}\
	else\
	{
		set state "nc"
	}
	if {[keylget TODO_REC $id.priority {}]}\
	{
		set priority [keylget TODO_REC $id.priority]
	}\
	else\
	{
		set priority 1
	}
	text .ask.description -width 30 -height 1 -wrap word
	pack .ask.description -side top -fill x -anchor w
	frame .ask.state
	pack .ask.state -side top
		radiobutton .ask.state.c -text "Completed" -variable state -value "c"
		pack .ask.state.c -side left
		radiobutton .ask.state.nc -text "Not completed" -variable state -value "nc"
		pack .ask.state.nc -side left
		radiobutton .ask.state.a -text "Aborted" -variable state -value "a"
		pack .ask.state.a -side left
	frame .ask.priority
	pack .ask.priority -side top
		label .ask.priority.lb -text "Priority "
		pack .ask.priority.lb -side left
		radiobutton .ask.priority.p1 -background $TODO_COL(1) -variable priority -value 1
		pack .ask.priority.p1 -side left
		radiobutton .ask.priority.p2 -background $TODO_COL(2) -variable priority -value 2
		pack .ask.priority.p2 -side left
		radiobutton .ask.priority.p3 -background $TODO_COL(3) -variable priority -value 3
		pack .ask.priority.p3 -side left
		radiobutton .ask.priority.p4 -background $TODO_COL(4) -variable priority -value 4
		pack .ask.priority.p4 -side left
		radiobutton .ask.priority.p5 -background $TODO_COL(5) -variable priority -value 5
		pack .ask.priority.p5 -side left
	frame .ask.b
	pack .ask.b -side bottom
		button .ask.b.ok -text Ok -width 10 \
			-command {
				set description "[.ask.description get 1.0 end-1c]"
				destroy .ask
			}
		pack .ask.b.ok -side left
		button .ask.b.cancel -text Cancel -width 10
		pack .ask.b.cancel -side left
		button .ask.b.del -text Delete -width 10 \
			-command "
				set description \"\"
				todo_del_record $org_id $proj_id $todo_id $id
				destroy .ask
			"
		pack .ask.b.del -side left
	if {[keylget TODO_REC $id.description {}]}\
	{
		.ask.description insert 1.0 "[keylget TODO_REC $id.description]"
		.ask.b.cancel configure -command "
			set priority $priority
			set state $state
			set description [list [keylget TODO_REC $id.description]]
			destroy .ask
		"
	}\
	else\
	{
		.ask.b.cancel configure -command "
			set priority $priority
			set state $state
			set description \"\"
			destroy .ask
		"
	}
	focus .ask.description
}

proc todo_popup_description {org_id proj_id todo_id id}\
{
	upvar #0 TODO_REC_${org_id}_${proj_id}_${todo_id} TODO_REC
	global description state priority TODO_COL
	todo_d_ask_description $org_id $proj_id $todo_id $id
	tkwait variable description
	if {$description == ""}\
	{
		if {[keylget TODO_REC $id {}]}\
		{
			keyldel TODO_REC $id
		}
		return
	}
	keylset TODO_REC $id.description "$description"
	keylset TODO_REC $id.state "$state"
	keylset TODO_REC $id.priority $priority
	.r.todo.c itemconfigure rect$id -fill $TODO_COL([keylget TODO_REC $id.priority])
	.r.todo.c itemconfigure desc$id -text "[keylget TODO_REC $id.description]"
	.r.todo.c itemconfigure st$id -text "[keylget TODO_REC $id.state]" -anchor nw
}

proc todo_new_record {org_id proj_id todo_id}\
{
	upvar #0 TODO_REC_${org_id}_${proj_id}_${todo_id} TODO_REC
	upvar #0 TOT_NUM_${org_id}_${proj_id}_${todo_id} TOT_NUM
	global description state priority CUR_Y TODO_COL
	set CUR_ID [expr [lindex [lsort [keylkeys TODO_REC]] end] + 1]
	todo_d_ask_description $org_id $proj_id $todo_id $CUR_ID
	tkwait variable description
	if {$description == ""}\
	{
		if {[keylget TODO_REC $CUR_ID {}]}\
		{
			keyldel TODO_DEL $CUR_ID
		}
		return
	}
	keylset TODO_REC $CUR_ID.description "$description"
	keylset TODO_REC $CUR_ID.state "$state"
	keylset TODO_REC $CUR_ID.priority $priority
	set tagrect [.r.todo.c create rectangle 10 [expr $CUR_Y - 5] 380 [expr $CUR_Y + 20] -fill $TODO_COL([keylget TODO_REC $CUR_ID.priority]) -outline {} -tags rect$CUR_ID]
	set tag1 [.r.todo.c create text 20 $CUR_Y -text "[keylget TODO_REC $CUR_ID.description]" -anchor nw -width 350 -tags desc$CUR_ID]
	set tag2 [.r.todo.c create text 350 $CUR_Y -text "[keylget TODO_REC $CUR_ID.state]" -anchor nw -tags st$CUR_ID]
	incr CUR_Y 30
	.r.todo.c bind $tagrect <1> "
		todo_popup_description $org_id $proj_id $todo_id $CUR_ID
	"
	.r.todo.c bind $tag1 <1> "
		todo_popup_description $org_id $proj_id $todo_id $CUR_ID
	"
	.r.todo.c bind $tag2 <1> "
		todo_popup_description $org_id $proj_id $todo_id $CUR_ID
	"
	.r.todo.c configure -scrollregion [.r.todo.c bbox all]
}

proc todo_del_record {org_id proj_id todo_id id}\
{
	upvar #0 TODO_REC_${org_id}_${proj_id}_${todo_id} TODO_REC
	if {! [keylget TODO_REC $id {}]}\
	{
		return
	}
	keyldel TODO_REC $id
	todo_create_list $org_id $proj_id $todo_id
}

#############################

proc todo_create_list {org_id proj_id todo_id}\
{
	upvar #0 TODO_REC_${org_id}_${proj_id}_${todo_id} TODO_REC
	global TODO_COL CUR_Y TODO_SORT
	.r.todo.c delete all
	.r.todo.c create rectangle 0 0 400 40 -fill #EEEEEE -outline {}
	set tag [.r.todo.c create text 20 20 -text "Description" -anchor nw]
	.r.todo.c bind $tag <1> "
		set TODO_SORT description
		todo_create_list $org_id $proj_id $todo_id
	"
	set tag [.r.todo.c create text 350 20 -text "State" -anchor n]
	.r.todo.c bind $tag <1> "
		set TODO_SORT state
		todo_create_list $org_id $proj_id $todo_id
	"
	set CUR_Y 50
	foreach i [lsort -command "todo_sort $org_id $proj_id $todo_id $TODO_SORT" [keylkeys TODO_REC]] \
	{
		set tagrect [.r.todo.c create rectangle 10 [expr $CUR_Y - 5] 380 [expr $CUR_Y + 20] -fill $TODO_COL([keylget TODO_REC $i.priority]) -outline {} -tags rect$i]
		set tag1 [.r.todo.c create text 20 $CUR_Y -text "[keylget TODO_REC $i.description]" -anchor nw -width 350 -tags desc$i]
		set tag2 [.r.todo.c create text 350 $CUR_Y -text "[keylget TODO_REC $i.state]" -anchor nw -tags st$i]
		incr CUR_Y 30
		.r.todo.c bind $tag1 <1> "
			todo_popup_description $org_id $proj_id $todo_id $i
		"
		.r.todo.c bind $tag2 <1> "
			todo_popup_description $org_id $proj_id $todo_id $i
		"
		.r.todo.c bind $tagrect <1> "
			todo_popup_description $org_id $proj_id $todo_id $i
		"
	}
	.r.todo.c configure -scrollregion [.r.todo.c bbox all]
}

